#
# pymldb
# Nicolas Kructhen, 2015-05-28
# Mich, 2016-01-26
# Copyright (c) 2013 Datacratic. All rights reserved.
#
from __future__ import absolute_import, division, print_function

from .version import __version__  # noqa

import pandas as pd
import requests
import json
from pymldb.util import add_repr_html_to_response
import threading
from .progress_monitor import ProgressMonitor


def decorate_response(fn):
    def inner(*args, **kwargs):
        result = add_repr_html_to_response(fn(*args, **kwargs))
        if result.status_code < 200 or result.status_code >= 400:
            raise ResourceError(result)
        return result
    return inner


class ResourceError(Exception):
    def __init__(self, r):
        try:
            message = json.dumps(r.json(), indent=2)
        except:
            message = r.content
        super(ResourceError, self).__init__(
            "'%d %s' response to '%s %s'\n\n%s" %
            (r.status_code, r.reason, r.request.method, r.request.url, message)
        )
        self.result = r


class Connection(object):

    def __init__(self, host="http://localhost", notebook=True):
        if not host.startswith("http"):
            raise Exception("URIs must start with 'http'")
        if host[-1] == '/':
            host = host[:-1]
        self.uri = host
        self.notebook = notebook

    @decorate_response
    def get(self, url, data=None, **kwargs):
        params = {}
        for k, v in kwargs.iteritems():
            if type(v) in [dict, list]:
                v = json.dumps(v)
            params[str(k)] = v
        return requests.get(self.uri + url, params=params, json=data)

    @decorate_response
    def put(self, url, payload=None):
        if payload is None:
            payload = {}
        return requests.put(self.uri + url, json=payload)

    @decorate_response
    def post(self, url, payload=None):
        if payload is None:
            payload = {}
        return requests.post(self.uri + url, json=payload)

    @decorate_response
    def delete(self, url):
        return requests.delete(self.uri + url)

    def query(self, sql, **kwargs):
        """
        Shortcut for GET /v1/query, except with argument format='dataframe'
        (the default), in which case it will simply wrap the result of the GET
        query to /v1/query (with format='table') in a `pandas.DataFrame`.
        """
        if 'format' not in kwargs or kwargs['format'] == 'dataframe':
            resp = self.get('/v1/query', data={'q': sql, 'format': 'table'}).json()
            if len(resp) == 0:
                return pd.DataFrame()
            else:
                return pd.DataFrame.from_records(resp[1:], columns=resp[0],
                                                 index="_rowName")
        kwargs['q'] = sql
        return self.get('/v1/query', **kwargs).json()

    def put_and_track(self, url, payload, refresh_rate_sec=1):
        """
        Put and track progress, displaying progress bars.

        May display the wrong progress if 2 things post/put on the same
        procedure name at the same time.
        """
        if not url.startswith('/v1/procedures'):
            raise Exception("The only supported route is /v1/procedures")
        parts = url.split('/')
        len_parts = len(parts)
        if len_parts not in [4, 6]:
            raise Exception(
                "You must either PUT a procedure or a procedure run")

        proc_id = parts[3]
        run_id = None

        if len_parts == 4:
                if 'params' not in payload:
                    payload['params'] = {}
                payload['params']['runOnCreation'] = True
        elif len_parts == 6:
            run_id = parts[-1]

        pm = ProgressMonitor(self, refresh_rate_sec, proc_id, run_id,
                             self.notebook)
        t = threading.Thread(target=pm.monitor_progress)
        t.start()

        try:
            return self.put(url, payload)
        except Exception as e:
            print(e)
        finally:
            pass
            pm.event.set()
            t.join()

    def post_and_track(self, url, payload, refresh_rate_sec=1):
        """
        Post and track progress, displaying progress bars.

        May display the wrong progress if 2 things post/put on the same
        procedure name at the same time.
        """
        if not url.startswith('/v1/procedures'):
            raise Exception("The only supported route is /v1/procedures")
        if url.endswith('/runs'):
            raise Exception(
                "Posting and tracking run is unsupported at the moment")
        if len(url.split('/')) != 3:
            raise Exception("You must POST a procedure")

        if 'params' not in payload:
            payload['params'] = {}
        payload['params']['runOnCreation'] = False

        res = self.post('/v1/procedures', payload).json()
        proc_id = res['id']

        pm = ProgressMonitor(self, refresh_rate_sec, proc_id,
                             notebook=self.notebook)

        t = threading.Thread(target=pm.monitor_progress)
        t.start()

        try:
            return self.post('/v1/procedures/{}/runs'.format(proc_id), {})
        except Exception as e:
            print(e)
        finally:
            pm.event.set()
            t.join()



"""
A few tests to make sure everything is in order.
usage: python pymldb/test.py <your_mldb_host>
"""
import sys
import pandas
import pymldb

host = sys.argv[1]

mldb = pymldb.Connection('http://' + host)

# create a dataset
mldb.put('/v1/datasets/patate', {'type': 'tabular'})
mldb.post('/v1/datasets/patate/rows', {
    'rowName': 'row0',
    'columns': [['x', 1, 0], ['y', 2, 0]]
})
mldb.post('/v1/datasets/patate/rows', {
    'rowName': 'row1',
    'columns': [['x', 2, 0], ['y', 3, 0]]
})
mldb.post('/v1/datasets/patate/commit')

# test mldb.query
# default format (dataframe)
df = pandas.DataFrame.from_records(
    [['row0', 1,2], ['row1', 2,3]],
    columns=['_rowName', 'x', 'y'],
    index='_rowName')
assert mldb.query('select * from patate order by x').equals(df)

# some other format
assert mldb.query('select * from patate order by x',
           format='soa',
           rowNames=False) == {'x': [1,2], 'y': [2,3]}

print('success!')



#
# steps_logger.py
# Mich, 2016-10-27
# Copyright (c) 2016 Datacratic. All rights reserved.
#
# Manages the display of steps.
#
from __future__ import absolute_import, division, print_function

from tqdm import tqdm, tqdm_notebook

def getStepsLogger(notebook):
    if notebook:
        return NotebookStepsLogger()
    return TextStepsLogger()

class TextStepsLogger(object):
    """
    Uses a single textual progress bar. Usefull for cli.
    """
    def __init__(self):
        self.steps = {}
        self.pbar = None

    def init_progress(self, steps_update):
        for step_update in steps_update:
            self.steps[step_update['name']] = {
                'done' : False
            }

    def log_progress_steps(self, steps_update):
        if len(self.steps) == 0:
            self.init_progress(steps_update)

        for step_update in steps_update:
            step = self.steps[step_update['name']]
            if 'ended' in step_update:
                if step['done']:
                    # already done, skip
                    continue

                step['done'] = True

                if self.pbar is None:
                    # if this step was too fast, there is no pbar yet for it
                    self.pbar = tqdm(desc=step_update['name'],
                                     unit='items',
                                     unit_scale=True)

                self.pbar.n = step_update['value'];
                self.pbar.update(0)
                self.pbar.close()
                self.pbar = None
            elif 'started' in step_update:
                if self.pbar is None:
                    self.pbar = tqdm(desc=step_update['name'],
                                unit='items',
                                unit_scale=True)
                self.pbar.n = step_update['value'];
                self.pbar.update(0)

    def clean_finish(self):
        self.steps = {}


class NotebookStepsLogger(object):
    """
    Displays notebook HTML progress bars for all the steps at once.
    """

    def __init__(self):
        self.steps = {}

    def __del__(self):
        for _, step in self.steps.iteritems():
            if not step['done']:
                step['pbar'].sp(bar_style='danger')
                step['pbar'].close()

    def init_progress(self, steps_update):
        for step_update in steps_update:
            self.steps[step_update['name']] = {
                'pbar' : tqdm_notebook(desc=step_update['name'],
                                       unit='items',
                                       unit_scale=True),
                'done' : False
            }

    def log_progress_steps(self, steps_update):
        if len(self.steps) == 0:
            self.init_progress(steps_update)

        for step_update in steps_update:
            step = self.steps[step_update['name']]
            pbar = step['pbar']
            if 'ended' in step_update:
                if step['done']:
                    # already done, skip
                    continue


                step['done'] = True
                pbar.n = step_update['value'];
                pbar.update(0)
                pbar.close()
            elif 'started' in step_update:
                pbar.n = step_update['value'];
                pbar.update(0)

    def clean_finish(self):
        for _, step in self.steps.iteritems():
            if not step['done']:
                step['pbar'].close()
        self.steps = {}

class StepsLogger(object):

    def __init__(self):
        self.steps = {}

    def init_progress(self, steps_update):
        for step_update in steps_update:
            self.steps[step_update['name']] = {
                'pbar' : None,
                'done' : False
            }

    def log_progress_steps(self, steps_update):
        if len(self.steps) == 0:
            self.init_progress(steps_update)

        for step_update in steps_update:
            step = self.steps[step_update['name']]
            if 'ended' in step_update:
                if step['done']:
                    # already done, skio
                    continue


                step['done'] = True
                step['pbar'].n = step_update['value'];
                step['pbar'].update(0) #step_update['value'] - step['pbar'])
                step['pbar'].close()
            elif 'started' in step_update:
                if step['pbar'] is None:
                    step['pbar'] = tqdm(desc=step_update['name'],
                                        unit='items',
                                        unit_scale=True)
                step['pbar'].n = step_update['value'];
                step['pbar'].update(0)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Datacratic Inc.  All rights reserved.
# @Author: Alexis Tremblay
# @Email: atremblay@datacratic.com
# @Date:   2015-01-07 15:45:01
# @Last Modified by:   Alexis Tremblay
# @Last Modified time: 2015-06-02 08:32:09
# @File Name:          data.py


import pandas as pd
from pymldb.query import Query
from pymldb.index import Time, Index
import requests
import logging
import numpy as np
logging.basicConfig(level=logging.DEBUG)


class BatFrame(object):
    def __init__(self, dataset_url):
        self.dataset_url = dataset_url
        self.query = Query(dataset_url)
        self._time = Time(dataset_url)
        self._index = Index(self)

    def __getitem__(self, val):

        if isinstance(val, str):
            col = Column(val, self.dataset_url)
            col.query.mergeQuery(self.query)
            return col
        elif isinstance(val, Query):
            bf = self.copy()
            # bf.query.addSELECT('*')
            bf.query.mergeQuery(val)
            return bf
        elif isinstance(val, slice):
            start = val.start
            stop = val.stop
            # step = val.step
            bf = self.copy()
            # bf.query.addSELECT('*')
            if start is not None:
                bf.query.setOFFSET(start)
            if stop is not None:
                bf.query.setLIMIT(stop)
            return bf
        elif isinstance(val, list):
            bf = self.copy()
            for value in val:
                bf.query.addSELECT("\"{}\"".format(value))
            return bf
        elif isinstance(val, Column):
            bf = self.copy()
            bf.query.addWHERE("({})".format(val.execution_name))
            return bf

    @property
    def columns(self):
        """Returns a numpy array of the columns name"""
        return requests.get(self.dataset_url + '/columns').json()

    @property
    def rows(self):
        """Returns a numpy array of the rows name"""
        bf = self.copy()
        result = bf.query.executeQuery(format="soa")
        return result["_rowName"]

    @property
    def time(self):
        copy_time = self._time.copy()
        return copy_time.query.mergeQuery(self.Query)

    @property
    def ix(self):
        copy_index = self._index.copy()
        return copy_index

    def copy(self):
        bf = BatFrame(self.dataset_url)
        bf.query = self.query.copy()
        return bf

    def toPandas(self):
        result = self.query.executeQuery(format="aos")
        if len(result) == 0:
            return pd.DataFrame()
        return pd.DataFrame.from_records(result, index="_rowName")

    def head(self, num_rows=5):
        bf = self.copy()
        bf.query.setLIMIT(num_rows)
        return bf

    def query(self, query):
        raise NotImplementedError()

    def sort(self, value, ascending=True):
        bf = self.copy()
        if not isinstance(value, list):
            value = [value]

        if not isinstance(ascending, list):
            ascending = [ascending]*len(value)

        if len(value) != len(ascending):
            raise RuntimeError("len(value) != len(ascending)")

        for by, asc in zip(value, ascending):
            if asc:
                sort = "ASC"
            else:
                sort = "DESC"
            bf.query.addORDERBY("\"{}\" {}".format(by, sort))
        return bf

    @property
    def shape(self):
        """
        Returns (rowCount, valueCount)
        """
        bf = self.copy()
        content = requests.get(bf.dataset_url).json()
        rowCount = content['status']['rowCount']
        valueCount = content['status']['valueCount']

        return (rowCount, valueCount)

    def __repr__(self):
        bf = self.copy()
        bf.query.setLIMIT(40)
        print(bf.toPandas())
        response = requests.get(bf.dataset_url).json()
        try:
            rowCount = response['status']['rowCount']
        except:
            rowCount = None

        if rowCount is not None and rowCount > 40:
            print("{} rows".format(rowCount))
        return ""


class Column(object):

    def __init__(self, name, dataset_url):
        """
        Parameters
        ----------
        name: string
            Name of the column. No check is actually done to see if the column
            exists.
        dataset_id:
            The base url where the dataset is located.
            e.g. localhost:8888/v1/datasets/<dataset_name>
        """
        logging.debug("Instanciating Column with {}".format(name))
        self.name = "\"{}\"".format(name)
        self.execution_name = "\"{}\"".format(name)
        self.dataset_url = dataset_url
        self.query = Query(dataset_url)
        self.query.addSELECT(self.name)

    @property
    def values(self):
        result = self.query.executeQuery(format="soa")
        if len(result) > 2:
            raise RuntimeError("Only one column should be returned")
        colName = [x for x in result.keys() if x != "_rowName"][0]
        return np.array(result[colName])

    def __getitem__(self, val):

        if isinstance(val, slice):
            start = val.start
            stop = val.stop
            # step = val.step
            col = self.copy()
            if start is not None:
                col.query.setOFFSET(start)
            if stop is not None:
                col.query.setLIMIT(stop)
            return col
        elif isinstance(val, Query):
            col = self.copy()
            col.query.mergeQuery(val)
            return col
        elif isinstance(val, str):
            col = self.copy()
            col.query.addWHERE("(rowName()='{}')".format(val))
            return col


    ####################
    #  Rich comparison  #
    ####################
    def _comparison(self, value, operator):
        """
        Parameters
        ----------
        value: Column object or base type
            The value against which to compare the column. It can either be
            another column or a base type value (e.g. int)

        Returns
        -------
        self.query

        Notes
        -----
        Returning self.query will allow the next object to use this column
        ops and concatenate something else
        """
        if isinstance(value, Column):
            self.query.addWHERE("(({}){}({}))".format(
                self.execution_name,
                operator,
                value.execution_name))
        elif isinstance(value, str):
            self.query.addWHERE("(({}){}\'{}\')".format(
                self.execution_name,
                operator,
                value))
        else:
            self.query.addWHERE("(({}){}({}))".format(
                self.execution_name,
                operator,
                value))

        copy = self.copy()
        copy.query.removeSELECT("{}".format(copy.execution_name))
        return copy.query

    def __eq__(self, value):
        return self._comparison(value, '=')

    def __ne__(self, value):
        return self._comparison(value, '!=')

    def __gt__(self, value):
        return self._comparison(value, '>')

    def __ge__(self, value):
        return self._comparison(value, '>=')

    def __lt__(self, value):
        return self._comparison(value, '<')

    def __le__(self, value):
        return self._comparison(value, '<=')

    ##################################
    #  Binary arithmetic operations  #
    ##################################
    def _binary_arithemtic(self, left, binary, right):
        """
        Parameters
        ----------
        operand: Column object, integer or float
            Value on which to apply operator to this column
        binary: char
            binary arithmetic operator (-, +, *, /, ^, %)

        Returns
        -------
        self

        Notes
        -----
        Returning self will allow the next object to use this column ops and
        concatenate something else
        """
        if isinstance(right, (int, float)):
            right = right
        elif isinstance(right, Column):
            right = right.execution_name
        else:
            raise AttributeError(
                "{} can only be used ".format(binary)
                + "with integer, float or column")

        if isinstance(left, (int, float)):
            left = left
        elif isinstance(left, Column):
            left = left.execution_name
        else:
            raise AttributeError(
                "{} can only be used ".format(binary)
                + "with integer, float or column")

        copy = self.copy()
        copy.query.removeSELECT("{}".format(copy.execution_name))
        if binary == '^':  # POWER needs a different treatment
            copy.execution_name = "pow({},{})".format(left, right)
        else:
            copy.execution_name = "{}{}{}".format(left, binary, right)
        copy.query.addSELECT(copy.execution_name)

        return copy

    def __mul__(self, value):
        return self._binary_arithemtic(self, '*', value)

    def __rmul__(self, value):
        return self._binary_arithemtic(value, '*', self)

    def __div__(self, value):
        if isinstance(value, (int, float)) and value == 0:
            raise ValueError(
                "Cannot divide by zero. "
                "Do you really want to explode the planet?")
        return self._binary_arithemtic(self, '/', value)

    def __rdiv__(self, value):
        return self._binary_arithemtic(value, '/', self)

    def __truediv__(self, value):
        if isinstance(value, (int, float)) and value == 0:
            raise ValueError(
                "Cannot divide by zero. "
                "Do you really want to explode the planet?")
        return self._binary_arithemtic(self, '/', value)

    def __rtruediv__(self, value):
        return self._binary_arithemtic(value, '/', self)

    def __sub__(self, value):
        return self._binary_arithemtic(self, '-', value)

    def __rsub__(self, value):
        return self._binary_arithemtic(value, '-', self)

    def __add__(self, value):
        return self._binary_arithemtic(self, '+', value)

    def __radd__(self, value):
        return self._binary_arithemtic(value, '+', self)

    def __pow__(self, value):
        return self._binary_arithemtic(self, '^', value)

    def __rpow__(self, value):
        return self._binary_arithemtic(value, '^', self)

    def __mod__(self, value):
        return self._binary_arithemtic(self, '%', value)

    def __rmod__(self, value):
        return self._binary_arithemtic(value, '%', self)

    def __or__(self, value):
        col = self.copy()
        left = self.execution_name
        right = value

        col.query.removeSELECT(left)
        if isinstance(right, Column):
            right = value.execution_name
            col.query.removeSELECT(right)
        elif isinstance(right, Query):
            right = right.WHERE

        col.query.addWHERE('(({}) OR ({}))'.format(left, right))
        return col.query

    def __and__(self, value):
        col = self.copy()
        left = self.execution_name
        right = value

        col.query.removeSELECT(left)
        if isinstance(right, Column):
            right = value.execution_name
            col.query.removeSELECT(right)
        elif isinstance(right, Query):
            right = right.WHERE

        col.query.addWHERE('(({}) AND ({}))'.format(left, right))

        return col.query

    def __rand__(self, value):
        col = self.copy()
        left = self.execution_name
        right = value

        col.query.removeSELECT(left)
        if isinstance(right, Column):
            right = value.execution_name
            col.query.removeSELECT(right)
        elif isinstance(right, Query):
            right = right.WHERE

        col.query.addWHERE('(({}) AND ({}))'.format(right, left))

    def __ror__(self, value):
        col = self.copy()
        left = self.execution_name
        right = value

        col.query.removeSELECT(left)
        if isinstance(right, Column):
            right = value.execution_name
            col.query.removeSELECT(right)
        elif isinstance(right, Query):
            right = right.WHERE

        col.query.addWHERE('(({}) OR ({}))'.format(right, left))
        return col.query

    #################################
    #  Unary arithmetic operations  #
    #################################
    def _unary_arithmetic(self, unary):
        """
        Parameters
        ----------
        unary: char
            Unary arithmetic operator (-, +) to be applied to this column

        Returns
        -------
        self

        Notes
        -----
        Returning self will allow the next object to use this column ops and
        concatenate something else
        """
        copy = self.copy()
        copy.query.removeSELECT("{}".format(copy.execution_name))
        copy.execution_name = "{}({})".format(unary, self.execution_name)
        copy.query.addSELECT(copy.execution_name)

        return copy

    def __neg__(self):
        return self._unary_arithmetic('-')

    def __pos__(self):
        raise NotImplementedError()

    def __invert__(self):
        copy = self.copy()
        copy.execution_name = "NOT {}".format(copy.execution_name)
        return copy

    def __abs__(self):
        raise NotImplementedError()

    #############
    #  Casting  #
    #############
    def __float__(self):
        raise NotImplementedError()

    def __int__(self):
        raise NotImplementedError()

    def __long__(self):
        raise NotImplementedError()

    ###########
    #  Other  #
    ###########

    def __iter__(self):
        result = self.query.executeQuery(format="soa")
        if len(result) > 2:
            raise RuntimeError("Only one column should be returned")
        colName = [x for x in result.keys() if x != "_rowName"][0]
        values = result[colName]

        i = 0
        while i < len(values):
            yield values[i]
            i += 1

    def max(self):
        copy = self.copy()
        copy.query.removeSELECT("{}".format(copy.execution_name))
        copy.execution_name = "max({})".format(self.execution_name)
        copy.query.addSELECT(copy.execution_name)
        copy.query.addGROUPBY(1)

        result = copy.query.executeQuery(format="table")
        return result[1][1]

    def min(self):
        copy = self.copy()
        copy.query.removeSELECT("{}".format(copy.execution_name))
        copy.execution_name = "min({})".format(self.execution_name)
        copy.query.addSELECT(copy.execution_name)
        copy.query.addGROUPBY(1)

        result = copy.query.executeQuery(format="table")
        return result[1][1]

    def copy(self):
        name = self.name[1:-1]  # Removing the surrounding ''
        col = Column(name, self.dataset_url)
        col.execution_name = self.execution_name
        col.query = self.query.copy()
        return col

    def count(self):
        """Return number of non-NA/null observations in the Series"""
        raise NotImplementedError()

    def head(self, n=5):
        """Returns first n rows"""
        col = self.copy()
        col.query.setLIMIT(n)
        return col.toPandas()

    def isnull(self):
        raise NotImplementedError()

    def isin(self, values):
        raise NotImplementedError()

    def value_counts(self):
        raise NotImplementedError()

    def unique(self):
        if self.name == self.execution_name:
            url = self.dataset_url + '/columns/{}/values'.format(
                self.name[1:-1])
            logging.debug("Getting values at {}".format(url))
            return requests.get(url).json()
        else:
            result = self.query.executeQuery(format="soa")
            if len(result) > 2:
                raise RuntimeError("Only one column should be returned")
            colName = [x for x in result.keys() if x != "_rowName"][0]
            return set(result[colName])

    def sort(self, ascending=True):
        col = self.copy()

        if ascending:
            sort = "ASC"
        else:
            sort = "DESC"
        col.query.addORDERBY("{} {}".format(col.execution_name, sort))
        return col

    def toPandas(self):
        result = self.query.executeQuery(format="soa")
        if len(result) > 2:
            raise RuntimeError("Only one column should be returned")
        colName = [x for x in result.keys() if x != "_rowName"][0]
        values = result[colName]
        rowName = result["_rowName"]
        if len(values) > 0:
            s = pd.Series(values, index=rowName)
        else:
            s = pd.Series()
        return s

    def __repr__(self):
        col = self.copy()
        col.query.setLIMIT(40)
        print(col.toPandas())
        response = requests.get(col.dataset_url).json()
        try:
            rowCount = response['status']['rowCount']
        except:
            rowCount = None

        if rowCount is not None and rowCount > 40:
            print("{} rows".format(rowCount))
        return ""



#
# progress_monitor.py
# Mich, 2017-01-12
# Copyright (c) 2017 Datacratic. All rights reserved.
#
# Tracks the progression of a procedure as long as it is running and updates
# progress output via a StepLogger. When in notebook mode, also displays a
# cancel button. The button # will fail if the notebook is not under the same
# host:port than MLDB because of cross origin policies.
#
from __future__ import absolute_import, division, print_function
import threading
import requests
from .steps_logger import getStepsLogger

class ProgressMonitor(object):

    def __init__(self, conn, refresh_rate_sec, proc_id, run_id=None,
                 notebook=True):
        self.conn = conn
        self.refresh_rate_sec = refresh_rate_sec
        self.proc_id = proc_id
        self.run_id = run_id
        self.event = threading.Event()
        self.notebook = notebook
        if self.notebook:
            from IPython.display import display, HTML
            self.display_html = lambda x: display(HTML(x))

    def monitor_progress(self):
        # wrap everything in a try/except because exceptions are not passed to
        # mldb.log by themselves.
        proc_id = self.proc_id
        run_id = self.run_id
        conn = self.conn
        refresh_rate_sec = 0.5
        run_id_flat = None
        try:
            # find run id
            sl = getStepsLogger(self.notebook)
            while not self.event.wait(refresh_rate_sec):
                if run_id is None:
                    res = conn.get('/v1/procedures/{}/runs'.format(proc_id)).json()
                    if res:
                        run_id = res[0]
                    else:
                        continue
                refresh_rate_sec = self.refresh_rate_sec
                if run_id_flat is None:
                    run_id_flat = run_id
                    for c in '-.:':
                        run_id_flat = run_id_flat.replace(c, '_')
                    if conn.uri == 'localhost':
                        host = ''
                    else:
                        host = conn.uri
                    if self.notebook:
                        self.display_html("""
                            <script type="text/javascript">
                                function cancel_{run_id_flat}(btn) {{
                                    $(btn).attr("disabled", true).html("Cancelling...");
                                    $.ajax({{
                                        url: "{host}/v1/procedures/{proc_id}/runs/{run_id}/state",
                                        type: 'PUT',
                                        data: JSON.stringify({{"state" : "cancelled"}}),
                                        success: () => {{ $(btn).html("Cancelled"); }},
                                        error: (xhr) => {{ console.error(xhr);
                                                            console.warn("If this is a Cross-Origin Request, this is a normal error. Otherwise you may report it.");
                                                            $(btn).html("Cancellation failed - See JavaScript console");
                                                        }}
                                    }});
                                }}
                            </script>
                            <button id="{run_id_flat}" onclick="cancel_{run_id_flat}(this);">Cancel</button>
                        """.format(run_id=run_id, run_id_flat=run_id_flat, proc_id=proc_id, host=host))
                res = requests.get(conn.uri + '/v1/procedures/{}/runs/{}'.format(proc_id, run_id)).json()
                if res['state'] == 'executing':
                    if 'steps' in res['progress']:
                        sl.log_progress_steps(res['progress']['steps'])
                    else:
                        sl.log_progress_steps([res['progress']])
                elif res['state'] == 'initializing':
                    continue
                else:
                    break
            if run_id is not None:
                if self.notebook:
                    self.display_html("""
                        <script type="text/javascript" class="removeMe">
                            $(function() {{
                                var outputArea = $(".removeMe").parents(".output_area:first");
                                outputArea.prevAll().remove();
                                outputArea.next().remove();
                                outputArea.remove();
                            }})
                        </script>
                    """.format(run_id_flat=run_id_flat))
                res = requests.get(conn.uri + '/v1/procedures/{}/runs/{}'.format(proc_id, run_id)).json()
                if res['state'] == 'finished':
                    sl.clean_finish()

        except Exception as e:
            print(str(e))
            import traceback
            print(traceback.format_exc())



#
# pymldb
# Copyright (c) 2013 Datacratic. All rights reserved.
#

import types, json
from pygments import highlight
from pygments.lexers import JsonLexer
from pygments.formatters import HtmlFormatter
from pygments.style import Style
from pygments.token import Keyword, Name, String, Number

class PygmentsStyle(Style):
    default_style = ""
    styles = {
        Name: 'bold #333', String: '#00d', Number: '#00d', Keyword: '#00d'
    }
        
def add_repr_html_to_response(resp):
    def _repr_html_(self):
        result = "<strong>%s %s</strong><br />" % (self.request.method, self.request.url)
        color = "black"
        if self.status_code >= 400: color = "red" 
        if self.status_code < 300: color = "green" 
        result += "<strong style=\"color: %s;\">%d %s</strong><br /> " % (color, self.status_code, self.reason)
        if "content-type" in self.headers:
            if self.headers["content-type"] == "application/json":
                result += highlight(
                    json.dumps(self.json(), indent=2), 
                    JsonLexer(), 
                    HtmlFormatter(noclasses = True, nobackground =True, style=PygmentsStyle)
                    )
            elif self.headers["content-type"] == "text/html":
                result += self.content
        return result
    resp._repr_html_ = types.MethodType(_repr_html_, resp)
    return resp




"""setuptools based setup module for pymldb.

based on:
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Defines __version__
exec(open('pymldb/version.py').read())

setup(
    name='pymldb',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    # https://www.python.org/dev/peps/pep-0440/#public-version-identifiers
    version=__version__,

    description='Python interface to MLDB',
    long_description=long_description,

    url='https://github.com/datacratic/pymldb',
    author='Datacratic',
    author_email='mldb@datacratic.com',
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',

        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',

        'Intended Audience :: Developers',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],

    keywords='machine learning',

    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'numpy',
        'pandas',
        'pygments',
        'requests[security]>=2.6',
    ],
)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Datacratic Inc.  All rights reserved.
# @Author:             Alexis Tremblay
# @Email:              atremblay@datacratic.com
# @Date:               2015-03-19 10:28:23
# @Last Modified by:   Alexis Tremblay
# @Last Modified time: 2015-04-21 09:05:54
# @File Name:          index.py

from pymldb.query import Query


class Time(object):
    """docstring for Time"""
    def __init__(self, dataset_url):
        super(Time, self).__init__()
        self.dataset_url = dataset_url
        self.query = Query(dataset_url)

    def __getitem__(self, value):
        print(value)

    def copy(self):
        copy_time = Time(self.dataset_url)
        copy_time = self.query.copy()
        return copy_time


class Index(object):
    """docstring for Index"""
    def __init__(self, bf):
        super(Index, self).__init__()
        self._bf = bf

    def copy(self):
        copy_index = Index(self._bf.copy())
        return copy_index

    def __getitem__(self, val):
        if isinstance(val, str):
            copy_bf = self._bf.copy()
            copy_bf.query.addWHERE("rowName()='{}'".format(val))
            return copy_bf
        elif isinstance(val, list):
            copy_bf = self._bf.copy()
            where = []
            for v in val:
                where.append("rowName()='{}'".format(v))

            copy_bf.query.addWHERE("({})".format(" OR ".join(where)))
            return copy_bf
        elif isinstance(val, tuple):
            if len(val) != 2:
                msg = "Too many indexers"
                raise IndexError(msg)

            row_index = val[0]
            col_index = val[1]
            row_index_bf = self._bf.ix[row_index]
            col_index_bf = row_index_bf[col_index]
            return col_index_bf
        else:
            raise NotImplementedError()



#
# magic.py
# Copyright (c) 2013 Datacratic. All rights reserved.
#

import requests, json
from pymldb.util import add_repr_html_to_response
import pandas as pd

host = "http://localhost"

###############################################################################
# Functions used by the cell and line magics

def print_usage_message(unknown = False):
    msg = """\
Usage:

  Line magic functions:

    %mldb help          
                        Print this message
    
    %mldb init <url>    
                        Initialize the plugins for the cell magics.
                        Extension comes pre-initialized with <uri> 
                        set to "http://localhost"
    
    %mldb doc <kind>/<type>    
                        Shows documentation in an iframe. <kind> can
                        be one of "datasets", "blocks", "pipelines" or
                        "plugins" and <type> can be one of the installed
                        types, e.g. pipelines/classifier. NB this will 
                        only work with an MLDB-hosted Notebook for now.

    %mldb query <sql>
                        Run an SQL-like query and return a pandas 
                        DataFrame. Dataset selection is done via the 
                        FROM clause.

    %mldb loadcsv <dataset> <url>
                        Create a dataset with id <dataset> from a CSV
                        hosted at the HTTP url <url>.
                        
    %mldb py <uri> <json args>
                        Run a python script named "main.py" from <uri>
                        and pass in <json args> as arguments.
                        <uri> can be one of:
                          - file://<rest of the uri>: a local directory
                          - gist://<rest of the uri>: a gist
                          - git://<rest of the uri>: a public git repo
                          - http(s)://<rest of the uri>: a file on the web

    %mldb pyplugin <name> <uri>
                        Load a python plugin called <name> from <uri> 
                        by executing its main.py. Any pre-existing plugin
                        called <name> will be deleted first.
                        <uri> can be one of:
                          - file://<rest of the uri>: a local directory
                          - gist://<rest of the uri>: a gist
                          - git://<rest of the uri>: a public git repo
                          - http(s)://<rest of the uri>: a file on the web
                          
    %mldb GET <route>
    %mldb DELETE <route>
                        HTTP GET/DELETE request to <route>. <route> should
                        start with a '/'.
                        
    %mldb GET <route> <json query params>       
                        HTTP GET request to <route>, JSON will be used to       
                        create query string. <route> should start with a '/'.       
                        
    %mldb PUT <route> <json>
    %mldb POST <route> <json>
                        HTTP PUT/POST request to <route>, <json> will
                        be sent as JSON payload. <route> should start
                        with a '/'.
       
                        
  Cell magic functions:

    %%mldb py <json args>
    <python code>
                        Run a python script in MLDB from the cell body.
    
    %%mldb query
    <sql>
                        Run an SQL-like query from the cell body and return
                        a pandas DataFrame. Dataset selection is done via
                        the FROM clause.
    
    %mldb loadcsv <dataset>
    <csv>
                        Create a dataset with id <dataset> from a CSV
                        in the cell body.
                        
    %%mldb GET <route>
    <json query params>
                        HTTP GET request to <route>, cell body will be
                        parsed as JSON and used to create query string.
                        <route> should start with a '/'.
                        
    %%mldb PUT <route>
    <json>
    %%mldb POST <route>
    <json>
                        HTTP PUT/POST request to <route>, cell body will
                        be sent as JSON payload. <route> should start
                        with a '/'.
"""
    if unknown:
        print("Unknown magic...")
        print("")
    print(msg)



def load_csv(dataset, csv_input):

    payload = {"args": [dataset, csv_input]}
    payload["source"] = """
import csv, urllib, StringIO
if mldb.script.args[1].startswith("http"):
    reader = csv.DictReader(open(urllib.urlretrieve(mldb.script.args[1])[0]))
else:
    reader = csv.DictReader(StringIO.StringIO(mldb.script.args[1]))

dataset = mldb.create_dataset(dict(id=mldb.script.args[0], type="beh.mutable"))
for i, row in enumerate(reader):
    values = []
    row_name = i
    for col in row:
        if col == "":
            row_name = row[col]
        else:
            values.append([col, row[col], 0])
    dataset.record_row(row_name, values)
dataset.commit()
print "Success!"
    """
    resp = requests.post(host+"/v1/types/plugins/python/routes/run",
                     data=json.dumps(payload))
    return handle_script_output(resp)

def handle_script_output(resp):
    if resp.status_code != 200:
        return add_repr_html_to_response(resp)
    
    result = resp.json()
    if "out" in result:
        for o in result["out"]:
            print(o[2])
    if "exception" in result:
        for e in result["exception"]["stack"]:
            print(e) 
    if "return" in result:
        return result["return"]

def run_query(q):
    global host

    resp = requests.get(host+"/v1/query", 
        data={"q": q, "format": "aos"})

    if resp.status_code != 200:
        return add_repr_html_to_response(resp)

    resp_json = resp.json()
    if len(resp_json) == 0: 
        return pd.DataFrame()
    else:
        return pd.DataFrame.from_records(resp_json, index="_rowName")


###############################################################################
# The Magic functions themselves


def dispatcher(line, cell=None):
    global host

    if line.strip() == "":
        return print_usage_message(True)
        
    
    parts = line.strip().split(" ")

    # The line magics
    if cell is None:

        # Init
        if len(parts) == 2 and parts[0] == "init":
            if not parts[1].startswith("http"):
                raise Exception("URI must start with 'http'")
            host = parts[1].strip("/")
            print("%mldb magic initialized with host as " + host)
            return

        # py or js: put a javascript or python script from an uri
        elif (len(parts) >= 2 and parts[0] in ["py", "js"]):

            type_name = "python"
            if parts[0] == "js":
                type_name = "javascript"
            payload = {"address": parts[1]}
            if len(parts) > 2:
                payload["args"] = json.loads(" ".join(parts[2:]))
            resp = requests.post(host+"/v1/types/plugins/" + type_name + "/routes/run",
                             data=json.dumps(payload))
            return handle_script_output(resp)
        
        # doc
        elif (len(parts) == 2 and parts[0] == "doc"):
            from IPython.display import IFrame
            return IFrame(src="/v1/types/"+parts[1]+"/doc", width=900, height=500)
        
        # pyplugin
        elif (len(parts) == 3 and parts[0] == "pyplugin"):
    
            name = parts[1]
            payload = {"type":"python", "params": {"address": parts[2]}}
            requests.delete(host+"/v1/plugins/" + name)
            resp = requests.put(host+"/v1/plugins/" + name,
                             data=json.dumps(payload))
            return add_repr_html_to_response(resp)

        # perform
        elif (len(parts) == 2 and parts[0] in ["GET", "DELETE"]):

            verb, uri = parts
            if verb == "GET":
                resp = requests.get(host+uri)
            elif verb == "DELETE":
                resp = requests.delete(host+uri)
                
            return add_repr_html_to_response(resp)

        # perform 
        elif (len(parts) > 2 and parts[0] in ["GET", "PUT", "POST"]):

            verb = parts[0]
            uri = parts[1]
            payload = json.loads(" ".join(parts[2:]))
            if verb == "GET":       
                for k in payload:       
                    if isinstance(payload[k], dict):        
                        payload[k] = json.dumps(payload[k])
                resp = requests.get(host+uri, params=payload)
            elif verb == "PUT":
                resp = requests.put(host+uri, data=json.dumps(payload))
            elif verb == "POST":
                resp = requests.post(host+uri, data=json.dumps(payload))
                
            return add_repr_html_to_response(resp)

        # help
        elif len(parts) == 1 and parts[0] == "help":
            return print_usage_message()

        elif (len(parts) > 1 and parts[0] == "query"):
            return run_query(" ".join(parts[1:]))
        
        elif (len(parts) == 3 and parts[0] == "loadcsv"):
            return load_csv(parts[1], parts[2])

        # We have something else
        else:
            return print_usage_message(True)

    # The cell magics
    else:
        # py or js: put a javascript or python script from an uri
        if (len(parts) >= 1 and parts[0] in ["py", "js"]):

            type_name = "python"
            if parts[0] == "js":
                type_name = "javascript"
            payload = {"source": cell}
            if len(parts) > 1:
                payload["args"] = json.loads(" ".join(parts[1:]))
            resp = requests.post(host+"/v1/types/plugins/" + type_name + "/routes/run",
                             data=json.dumps(payload))
            
            return handle_script_output(resp)
        
        if (len(parts) == 1 and parts[0] == "query"):
            return run_query(cell.replace("\n", " ").strip())

        if (len(parts) == 2 and parts[0] == "loadcsv"):
            return load_csv(parts[1], cell)

        # perform
        elif (len(parts) == 2 and parts[0] in ["GET", "PUT", "POST"]):

            verb, uri = parts
            payload = json.loads(cell)
            if verb == "GET":
                for k in payload:
                    if isinstance(payload[k], dict):
                        payload[k] = json.dumps(payload[k])
                        
                resp = requests.get(host+uri, params=payload)
            elif verb == "PUT":
                resp = requests.put(host+uri, data=json.dumps(payload))
            elif verb == "POST":
                resp = requests.post(host+uri, data=json.dumps(payload))
                
            return add_repr_html_to_response(resp)
        # help
        elif len(parts) == 1 and parts[0] == "help":
            return print_usage_message()
            

        # We have something else
        else:
            return print_usage_message(True)




# -*- coding: utf-8 -*-
# Copyright (c) 2015 Datacratic Inc.  All rights reserved.
# @Author:             Alexis Tremblay
# @Email:              atremblay@datacratic.com
# @Date:               2015-03-06 14:53:37
# @Last Modified by:   Alexis Tremblay
# @Last Modified time: 2015-04-09 16:54:58
# @File Name:          query.py


import json
import requests
import traceback
import copy
from collections import Counter
import logging
logging.basicConfig(level=logging.DEBUG)


class Query(object):
    """docstring for Query"""
    def __init__(self, dataset_url):
        self.dataset_url = dataset_url
        self.SELECT = Counter()
        self.WHERE = None # list()
        self.GROUPBY = list()
        self.OFFSET = None
        self.LIMIT = None
        self.ORDERBY = list()

    def addSELECT(self, obj):
        logging.debug("Adding SELECT {}".format(obj))
        self.SELECT[obj] += 1
        logging.debug(self.SELECT)

    def removeSELECT(self, obj):
        logging.debug("Removing SELECT {}".format(obj))
        if obj not in self.SELECT:
            return

        self.SELECT[obj] -= 1
        if self.SELECT[obj] == 0:
            del self.SELECT[obj]
        logging.debug(self.SELECT)

    def mergeSELECT(self, query):
        self.SELECT = self.SELECT + query.SELECT

    def addWHERE(self, where, boolean=None):
        if where is None:
            return

        if self.WHERE is None:
            self.WHERE = where
        else:
            # if boolean is None:
            #     raise RuntimeError("Must provide boolean instruction to WHERE")
            if boolean != "OR" and boolean != "AND":
                raise RuntimeError("Boolean instruction must OR or AND")
            self.WHERE = "({} {} {})".format(self.WHERE, boolean, where)

        # self.WHERE.append(where)

    def mergeWHERE(self, query, how):
        self.addWHERE(query.WHERE, how)
        # self.WHERE.extend(query.WHERE)

    def addGROUPBY(self, value):
        self.GROUPBY.append(str(value))

    def mergeGROUPBY(self, query):
        self.GROUPBY.extend(query.GROUPBY)

    def setOFFSET(self, value):
        # Basically the start of slicing. This can normally be a negative
        # number in python. For now (and probably forever), not supported.
        # i.e. my_list[-10:] is not supported
        if type(value) != int:
            raise RuntimeError("Can only slice with integer")
        if value < 0:
            raise RuntimeError("Slicing with negative index is not allowed")
        if self.OFFSET is None:
            self.OFFSET = value
        if self.OFFSET < value:
            self.OFFSET = value

    def setLIMIT(self, value):
        # Basically the stop of slicing. This can normally be a negative
        # number in python. For now (and probably forever), not supported.
        # i.e. my_list[:-1] is not supported
        if type(value) != int:
            raise RuntimeError("Can only slice with integer")
        if value < 0:
            raise RuntimeError("Slicing with negative index is not allowed")
        if self.LIMIT is None:
            self.LIMIT = value
        if self.LIMIT > value:
            self.LIMIT = value

    def addORDERBY(self, value):
        self.ORDERBY.append(value)

    def mergeORDERBY(self, query):
        self.ORDERBY.extend(query.ORDERBY)

    def mergeQuery(self, query, how=None):
        self.mergeSELECT(query)
        self.mergeWHERE(query, how)
        self.mergeGROUPBY(query)
        self.mergeORDERBY(query)

        if self.OFFSET is not None and query.OFFSET is not None:
            raise RuntimeError("Multiple slicing asked")

        if self.OFFSET is None:
            self.OFFSET = query.OFFSET

        if self.LIMIT is not None and query.LIMIT is not None:
            raise RuntimeError("Multiple slicing asked")

        if self.LIMIT is None:
            self.LIMIT = query.LIMIT

    def buildQuery(self):
        data = {}
        # print("Building query")
        # print(self.SELECT, len(self.SELECT))
        if len(self.SELECT) == 0:
            # print("Replacing SELECT with *")
            data["select"] = '*'
        else:
            data["select"] = ",".join(self.SELECT.keys())

        if self.WHERE is not None:
            # data["where"] = " ".join(self.WHERE)
            data["where"] = self.WHERE
        if len(self.GROUPBY) > 0:
            data["groupBy"] = ",".join(self.GROUPBY)
        if self.OFFSET is not None:
            data["offset"] = self.OFFSET
        if self.LIMIT is not None:
            data["limit"] = self.LIMIT
        if len(self.ORDERBY) > 0:
            data["orderBy"] = ",".join(self.ORDERBY)

        return data

    def executeQuery(self, format):

        query = self.buildQuery()
        query["format"] = format
        logging.debug("REST params\n{}".format(json.dumps(query)))

        select_url = self.dataset_url + "/query"

        try:
            # logging.info(select_url)
            response = requests.get(select_url, params=query)
            logging.info("URL poked {}".format(response.url))
        except requests.HTTPError as e:
            logging.error("Code: {}\nReason: {}".format(
                e.status_code, e.reason))
            logging.error("Content: {}".format(response.content))
            logging.error(traceback.format_exc())

        if response.status_code != 200:
            logging.error("Code: {}\nReason: {}".format(
                response.status_code, response.reason))
            logging.error("Content: {}".format(response.content))
            logging.error(traceback.format_exc())

        try:
            return response.json()
        except:
            return {}

    def __or__(self, value):
        if isinstance(value, Query):
            query = self.copy()
            # self.addWHERE('OR')
            query.mergeQuery(value, "OR")
        return query

    def __and__(self, value):
        if isinstance(value, Query):
            query = self.copy()
            # self.addWHERE('AND')
            query.mergeQuery(value, "AND")
        return query

    def __rand__(self, value):
        raise NotImplementedError()

    def __ror__(self, value):
        raise NotImplementedError()

    def copy(self):
        query = Query(self.dataset_url)
        query.SELECT = copy.deepcopy(self.SELECT)
        query.WHERE = copy.deepcopy(self.WHERE)
        query.ORDERBY = copy.deepcopy(self.ORDERBY)
        query.GROUPBY = copy.deepcopy(self.GROUPBY)
        query.OFFSET = copy.deepcopy(self.OFFSET)
        query.LIMIT = copy.deepcopy(self.LIMIT)
        return query

    def __repr__(self):
        return json.dumps(self.buildQuery(), indent=4)

    def __str__(self):
        return json.dumps(self.buildQuery(), indent=4)

